with Casendra.Cases;
with Casendra.Strata;
with Ada.Calendar.Formatting;
with Ada.Calendar.Arithmetic;
with Ada.Characters.Handling;
with Ada.Command_Line;
with Ada.Text_IO;
with Ada.Characters.Latin_1;
with Ada.Unchecked_Deallocation;
with Ada.Strings.Fixed;
with Ada.Strings.Unbounded; use Ada.Strings.Unbounded;
with Config_File;
with GNAT.OS_Lib;
with Casendra.Attachment;
with Casendra.Attachments;
with Casendra.Text_Utils;
with Sigwinch_Handler;
with Terminal_Manager;

procedure Csdownloader (Caseid      : in String := "";
                        Input       : in String := "";
                        Interactive :    Boolean) 
  -- TODO: Add simulate only option to save traffic
  -- Main procedure for downloading case attachments and/or comments
  -- it works interactively using shell and asks user for number of attachments to be downloaded or range of date
  -- All attachements are downloaded into directory specified in config
  
is
   
   procedure Usage is
      use Ada.Text_IO;
   begin
      Put_Line ("casendrda download <case number>");
      Gnat.OS_Lib.OS_Exit (1);
   end Usage;
   
   Working_On     : Casendra.Cases.Case_T;
   Finished       : Boolean         := False;
   Selection      : Casendra.Attachments.Attachments_T;
   Dir            : constant String := Config_FIle.Get_String (Casendra.Config,
                                                               "downloader.directory",
                                                               False, 
                                                               "/tmp/folder/");
   task Signal_Handler;
   
   task body Signal_Handler is
   begin
      while not Finished loop
         select
            Sigwinch_Handler.Handler.Wait;
            Terminal_Manager.Manager.Update;
         or
            delay 0.1;
            exit when Finished;
         end select;
      end loop;
   end Signal_Handler;
   
   procedure Display_Progress (Percents_Downloaded : in Natural;
                               Finished : Boolean)
   is
      Screen_Width      : constant Natural := Terminal_Manager.Manager.Get_Width;
      Minimal_String    : constant String  := "[]100%";
      Progress_Bar_Size : constant Natural := Screen_Width - Minimal_String'Length;
      Number_Of_Arrows  : Natural;
   begin
      if Screen_Width < Minimal_String'Length then
         return;
      end if;
      Number_Of_Arrows := Percents_Downloaded * Progress_Bar_Size / 100;
      Ada.Text_IO.Put(ASCII.ESC & "[2K" & ASCII.CR);
      declare
         Print    : String (1 .. Screen_Width);
         Percents : String := Natural'Image (Percents_Downloaded) & "%";
      begin
         Print (Print'First)                                    := '[';
         Print (Print'First + 1 .. Number_Of_Arrows)            := (others => '=');
         Print (Number_Of_Arrows + 1 .. Progress_Bar_Size)      := (others => '.');
         Print (Progress_Bar_Size + 1)                          := ']';
         Print (Progress_Bar_Size + 2 .. Print'Last)            := (others => ' ');
         Print (Print'Last - Percents'Length + 1 .. Print'last) := Percents;
         Ada.Text_Io.Put(Print);
      end;
      if Finished then
         Ada.Text_Io.New_Line;
      end if;
   end Display_Progress;
   
   Progress_Callback : Casendra.Strata.Progress_Callback_T := Display_Progress'Unrestricted_Access;
   
   function Select_Attachment_To_Download (Case_Object  : in Casendra.Cases.Case_T) return Casendra.Attachments.Attachments_T 
   is
      function Get_Time_Stamp (Raw_Selection    : in String) return Ada.Calendar.Time is
         Unit                           : Character             := Raw_Selection (Raw_Selection'Last);
         Day_Duration                   : constant Duration     := Ada.Calendar.Day_Duration'Last;
         Hour_Duration                  : constant Duration     := Ada.Calendar.Day_Duration'Last / 24;
         Dur                            : Duration;
         use Ada.Calendar;
      begin
         if Raw_Selection'Length < 2 or else
           (for some X of Raw_Selection (Raw_Selection'First .. Raw_Selection'Last - 1) => not Ada.Characters.Handling.Is_Digit (X)) then
            raise Constraint_Error with "Wrong time format"     ;
         end if;
         case Unit is
            when 'd'    =>
               Dur := Day_Duration * Natural'Value (Raw_Selection (Raw_Selection'First .. Raw_Selection'Last -1 ));
            when 'h' =>
               Dur := Hour_Duration * Natural'Value (Raw_Selection (Raw_Selection'First .. Raw_Selection'Last -1 ));
            when others =>
               Ada.Text_Io.Put_Line ("Only h (hours) and d (days) is allowed");
               raise Constraint_Error;
         end case;
         return  Clock - Dur;
      end Get_Time_Stamp;
      
      Attachments       : Casendra.Attachments.Attachments_T    := Case_Object.Get_All_Attachments;
      Counter           : Positive;
   begin
      Counter := 1;
      for Attachment of Attachments loop
         declare
            Days                : Ada.Calendar.Arithmetic.Day_Count;
            Seconds             : Duration;
            Leap_Seconds        : Ada.Calendar.Arithmetic.Leap_Seconds_Count;
         begin
            -- Fixme
            if not Attachment.Is_Downloaded (Dir & Working_On.Id & "/attachments/") then
               Ada.Calendar.Arithmetic.Difference (Right           => Attachment.Created_At,
                                                   Left            => Ada.Calendar.Clock, 
                                                   Days            => Days, 
                                                   Seconds         => Seconds, 
                                                   Leap_Seconds    => Leap_Seconds);
               if Interactive then
                  Ada.Text_Io.Put_Line (Counter'Img &
                                          "; Filename: " & To_String (Attachment.File_Name) & 
                                          "; Length = " & Natural'Image (Attachment.Length / 1024) & " Kb" & 
                                          "; Uploaded " & Days'Img & " days ago");
               end if;
            end if;
            Counter := Counter + 1;
         end;
      end loop;
      if Interactive then
         Ada.Text_IO.Put_Line ("Please enter attachment(-s) to download separated by comma (Hit ENTER to dowload all). ");
         Ada.Text_Io.Put_Line ("To download attchemts for last N days (hours) use Nd (or Nh):");
         Ada.Text_Io.Put ("To download last N attachments use -N :");
      end if;
      declare
         Raw_Selection  : constant String                               := (if not Interactive then Input else Ada.Text_IO.Get_Line);
         Delimeter      : constant String                               := ",";
         Retval         : Casendra.Attachments.Attachments_T;
      begin
         -- If empty string was provided we're going to download all attachments
         if Raw_Selection'Length = 0 then
            return Attachments;
         end if;
         if Ada.Characters.Handling.Is_Letter (Raw_Selection (Raw_Selection'Last)) then
            -- Assuming time delta has been specified
            declare 
               use Ada.Calendar;
               Timestamp        : Time  := Get_Time_Stamp (Raw_Selection);
            begin
               for Attach of Attachments  loop
                  if Timestamp < Attach.Created_At then
                     Retval.Append (Attach);
                  end if;
               end loop;
            end;
               
         else
            for Index of Casendra.Text_Utils.String_To_Natural_Array (Raw_Selection, Delimeter) loop
               Retval.Append (Attachments (Index));
            end loop;
         end if;
         return Retval;
      end;
   end Select_Attachment_To_Download;
   
   
   
   use type Casendra.Text_Utils.Natural_Array_Access;

begin
   if not Interactive then
      Casendra.Cases.Init (Working_On, Caseid);
   else
      if Ada.Command_Line.Argument_Count < 1 then
         Usage;
      end if;
      Casendra.Cases.Init (Working_On, Ada.Command_Line.Argument (1));
   end if;
   
   Casendra.Cases.Save_History (Working_On, Dir);
   loop
  Try:
      begin
         Selection      := Select_Attachment_To_Download (Working_On);
         exit;
      exception
         when others    =>
            if Interactive then
               Ada.Text_IO.Put_Line ("Invalid input. Please try again!");
            else
               raise;
            end if;
      end Try;
   end loop;
   
   if Interactive then
      Casendra.Cases.Download_Attachment (Working_On,
                                          Selection, 
                                          Dir      => Dir,
                                          Callback => Progress_Callback);
   else
      Casendra.Cases.Download_Attachment (Working_On,
                                          Selection, 
                                          Dir      => Dir,
                                          Callback => null);
   end if;
   
   
   Finished := True;
   
end Csdownloader;
